@@ -91,6 +91,25 @@ void resolver_shutdown(void)
 #endif
 }
 
+resolver_srv_rr_t *resolver_srv_rr_new(xmpp_ctx_t *ctx,
+                                       const char *host,
+                                       unsigned short port,
+                                       unsigned short prio,
+                                       unsigned short weight)
+{
+    resolver_srv_rr_t *rr = strophe_alloc(ctx, sizeof(*rr));
+    if (rr) {
+        memset(rr, 0, sizeof(*rr));
+        rr->port = port;
+        rr->priority = prio;
+        rr->weight = weight;
+        if (host) {
+            snprintf(rr->target, sizeof(rr->target), "%s", host);
+        }
+    }
+    return rr;
+}
+
 static void resolver_srv_list_sort(resolver_srv_rr_t **srv_rr_list)
 {
     resolver_srv_rr_t *rr_head;
@@ -158,6 +177,10 @@ int resolver_srv_lookup_buf(xmpp_ctx_t *ctx,
     set = resolver_ares_srv_lookup_buf(ctx, buf, len, srv_rr_list);
 #else
     set = resolver_raw_srv_lookup_buf(ctx, buf, len, srv_rr_list);
+    if (set != XMPP_DOMAIN_FOUND && *srv_rr_list != NULL) {
+        resolver_srv_free(ctx, *srv_rr_list);
+        *srv_rr_list = NULL;
+    }
 #endif
     resolver_srv_list_sort(srv_rr_list);
 
@@ -179,8 +202,8 @@ int resolver_srv_lookup(xmpp_ctx_t *ctx,
     (void)buf;
     (void)len;
 
-    xmpp_snprintf(fulldomain, sizeof(fulldomain), "_%s._%s.%s", service, proto,
-                  domain);
+    strophe_snprintf(fulldomain, sizeof(fulldomain), "_%s._%s.%s", service,
+                     proto, domain);
 
     *srv_rr_list = NULL;
 
@@ -196,7 +219,7 @@ int resolver_srv_lookup(xmpp_ctx_t *ctx,
         return set;
 #endif /* _WIN32 */
 
-    buf = xmpp_alloc(ctx, RESOLVER_BUF_MAX);
+    buf = strophe_alloc(ctx, RESOLVER_BUF_MAX);
     if (buf == NULL)
         return XMPP_DOMAIN_NOT_FOUND;
 
@@ -210,7 +233,7 @@ int resolver_srv_lookup(xmpp_ctx_t *ctx,
     if (len > 0)
         set = resolver_srv_lookup_buf(ctx, buf, (size_t)len, srv_rr_list);
 
-    xmpp_free(ctx, buf);
+    strophe_free(ctx, buf);
 
 #endif /* HAVE_CARES */
 
@@ -223,7 +246,7 @@ void resolver_srv_free(xmpp_ctx_t *ctx, resolver_srv_rr_t *srv_rr_list)
 
     while (srv_rr_list != NULL) {
         rr = srv_rr_list->next;
-        xmpp_free(ctx, srv_rr_list);
+        strophe_free(ctx, srv_rr_list);
         srv_rr_list = rr;
     }
 }
@@ -321,6 +344,9 @@ static unsigned message_name_get(const unsigned char *buf,
             if (i >= buf_len)
                 return 0;
             pointer = (label_len & 0x3f) << 8 | buf[i++];
+            /* Prevent infinite looping */
+            if (pointer >= buf_offset)
+                return 0;
             if (name != NULL && name_len >= name_max && name_max > 0) {
                 /* We have filled the name buffer. Don't pass it recursively. */
                 name[name_max - 1] = '\0';
@@ -431,17 +457,19 @@ static int resolver_raw_srv_lookup_buf(xmpp_ctx_t *ctx,
         rdlength = xmpp_ntohs_ptr(&buf[j + 8]);
         j += 10;
         if (type == MESSAGE_T_SRV && class == MESSAGE_C_IN) {
-            rr = xmpp_alloc(ctx, sizeof(*rr));
-            rr->next = *srv_rr_list;
-            rr->priority = xmpp_ntohs_ptr(&buf[j]);
-            rr->weight = xmpp_ntohs_ptr(&buf[j + 2]);
-            rr->port = xmpp_ntohs_ptr(&buf[j + 4]);
-            name_len = message_name_get(buf, len, j + 6, rr->target,
-                                        sizeof(rr->target));
-            if (name_len > 0)
-                *srv_rr_list = rr;
-            else
-                xmpp_free(ctx, rr); /* skip broken record */
+            rr = resolver_srv_rr_new(ctx, NULL, 0, 0, 0);
+            if (rr) {
+                rr->next = *srv_rr_list;
+                rr->priority = xmpp_ntohs_ptr(&buf[j]);
+                rr->weight = xmpp_ntohs_ptr(&buf[j + 2]);
+                rr->port = xmpp_ntohs_ptr(&buf[j + 4]);
+                name_len = message_name_get(buf, len, j + 6, rr->target,
+                                            sizeof(rr->target));
+                if (name_len > 0)
+                    *srv_rr_list = rr;
+                else
+                    strophe_free(ctx, rr); /* skip broken record */
+            }
         }
         j += rdlength;
     }
@@ -480,7 +508,7 @@ static int resolver_ares_srv_lookup_buf(xmpp_ctx_t *ctx,
 
     item = srv;
     while (item != NULL) {
-        rr = xmpp_alloc(ctx, sizeof(*rr));
+        rr = strophe_alloc(ctx, sizeof(*rr));
         if (rr == NULL)
             break;
         rr->next = *srv_rr_list;
@@ -705,15 +733,15 @@ static int resolver_win32_srv_lookup(xmpp_ctx_t *ctx,
 
                 while (current) {
                     if (current->wType == DNS_TYPE_SRV) {
-                        rr = xmpp_alloc(ctx, sizeof(*rr));
+                        rr = strophe_alloc(ctx, sizeof(*rr));
                         if (rr == NULL)
                             break;
                         rr->next = *srv_rr_list;
                         rr->port = current->Data.Srv.wPort;
                         rr->priority = current->Data.Srv.wPriority;
                         rr->weight = current->Data.Srv.wWeight;
-                        xmpp_snprintf(rr->target, sizeof(rr->target), "%s",
-                                      current->Data.Srv.pNameTarget);
+                        strophe_snprintf(rr->target, sizeof(rr->target), "%s",
+                                         current->Data.Srv.pNameTarget);
                         *srv_rr_list = rr;
                     }
                     current = current->pNext;
