@@ -39,7 +39,7 @@ xmpp_stanza_t *xmpp_stanza_new(xmpp_ctx_t *ctx)
 {
     xmpp_stanza_t *stanza;
 
-    stanza = xmpp_alloc(ctx, sizeof(xmpp_stanza_t));
+    stanza = strophe_alloc(ctx, sizeof(xmpp_stanza_t));
     if (stanza != NULL) {
         stanza->ref = 1;
         stanza->ctx = ctx;
@@ -124,7 +124,7 @@ xmpp_stanza_t *xmpp_stanza_copy(const xmpp_stanza_t *stanza)
     copy->type = stanza->type;
 
     if (stanza->data) {
-        copy->data = xmpp_strdup(stanza->ctx, stanza->data);
+        copy->data = strophe_strdup(stanza->ctx, stanza->data);
         if (!copy->data)
             goto copy_error;
     }
@@ -182,14 +182,15 @@ int xmpp_stanza_release(xmpp_stanza_t *stanza)
         while (child) {
             tchild = child;
             child = child->next;
+            tchild->next = NULL;
             xmpp_stanza_release(tchild);
         }
 
         if (stanza->attributes)
             hash_release(stanza->attributes);
         if (stanza->data)
-            xmpp_free(stanza->ctx, stanza->data);
-        xmpp_free(stanza->ctx, stanza);
+            strophe_free(stanza->ctx, stanza->data);
+        strophe_free(stanza->ctx, stanza);
         released = 1;
     }
 
@@ -263,7 +264,7 @@ static char *_escape_xml(xmpp_ctx_t *ctx, char *text)
             len++;
         }
     }
-    if ((buf = xmpp_alloc(ctx, (len + 1) * sizeof(char))) == NULL)
+    if ((buf = strophe_alloc(ctx, (len + 1) * sizeof(char))) == NULL)
         return NULL; /* Error */
     dst = buf;
     for (src = text; *src != '\0'; src++) {
@@ -336,8 +337,8 @@ _render_stanza_recursive(xmpp_stanza_t *stanza, char *buf, size_t buflen)
         tmp = _escape_xml(stanza->ctx, stanza->data);
         if (tmp == NULL)
             return XMPP_EMEM;
-        ret = xmpp_snprintf(ptr, left, "%s", tmp);
-        xmpp_free(stanza->ctx, tmp);
+        ret = strophe_snprintf(ptr, left, "%s", tmp);
+        strophe_free(stanza->ctx, tmp);
         if (ret < 0)
             return XMPP_EMEM;
         _render_update(&written, buflen, ret, &left, &ptr);
@@ -346,7 +347,7 @@ _render_stanza_recursive(xmpp_stanza_t *stanza, char *buf, size_t buflen)
             return XMPP_EINVOP;
 
         /* write beginning of tag and attributes */
-        ret = xmpp_snprintf(ptr, left, "<%s", stanza->data);
+        ret = strophe_snprintf(ptr, left, "<%s", stanza->data);
         if (ret < 0)
             return XMPP_EMEM;
         _render_update(&written, buflen, ret, &left, &ptr);
@@ -374,8 +375,8 @@ _render_stanza_recursive(xmpp_stanza_t *stanza, char *buf, size_t buflen)
                     hash_iter_release(iter);
                     return XMPP_EMEM;
                 }
-                ret = xmpp_snprintf(ptr, left, " %s=\"%s\"", key, tmp);
-                xmpp_free(stanza->ctx, tmp);
+                ret = strophe_snprintf(ptr, left, " %s=\"%s\"", key, tmp);
+                strophe_free(stanza->ctx, tmp);
                 if (ret < 0) {
                     hash_iter_release(iter);
                     return XMPP_EMEM;
@@ -387,7 +388,7 @@ _render_stanza_recursive(xmpp_stanza_t *stanza, char *buf, size_t buflen)
 
         if (!stanza->children) {
             /* write end if singleton tag */
-            ret = xmpp_snprintf(ptr, left, "/>");
+            ret = strophe_snprintf(ptr, left, "/>");
             if (ret < 0)
                 return XMPP_EMEM;
             _render_update(&written, buflen, ret, &left, &ptr);
@@ -395,7 +396,7 @@ _render_stanza_recursive(xmpp_stanza_t *stanza, char *buf, size_t buflen)
             /* this stanza has child stanzas */
 
             /* write end of start tag */
-            ret = xmpp_snprintf(ptr, left, ">");
+            ret = strophe_snprintf(ptr, left, ">");
             if (ret < 0)
                 return XMPP_EMEM;
             _render_update(&written, buflen, ret, &left, &ptr);
@@ -413,7 +414,7 @@ _render_stanza_recursive(xmpp_stanza_t *stanza, char *buf, size_t buflen)
             }
 
             /* write end tag */
-            ret = xmpp_snprintf(ptr, left, "</%s>", stanza->data);
+            ret = strophe_snprintf(ptr, left, "</%s>", stanza->data);
             if (ret < 0)
                 return XMPP_EMEM;
 
@@ -447,7 +448,7 @@ int xmpp_stanza_to_text(xmpp_stanza_t *stanza, char **buf, size_t *buflen)
 
     /* allocate a default sized buffer and attempt to render */
     length = 1024;
-    buffer = xmpp_alloc(stanza->ctx, length);
+    buffer = strophe_alloc(stanza->ctx, length);
     if (!buffer) {
         *buf = NULL;
         *buflen = 0;
@@ -456,16 +457,16 @@ int xmpp_stanza_to_text(xmpp_stanza_t *stanza, char **buf, size_t *buflen)
 
     ret = _render_stanza_recursive(stanza, buffer, length);
     if (ret < 0) {
-        xmpp_free(stanza->ctx, buffer);
+        strophe_free(stanza->ctx, buffer);
         *buf = NULL;
         *buflen = 0;
         return ret;
     }
 
     if ((size_t)ret > length - 1) {
-        tmp = xmpp_realloc(stanza->ctx, buffer, ret + 1);
+        tmp = strophe_realloc(stanza->ctx, buffer, ret + 1);
         if (!tmp) {
-            xmpp_free(stanza->ctx, buffer);
+            strophe_free(stanza->ctx, buffer);
             *buf = NULL;
             *buflen = 0;
             return XMPP_EMEM;
@@ -475,7 +476,7 @@ int xmpp_stanza_to_text(xmpp_stanza_t *stanza, char **buf, size_t *buflen)
 
         ret = _render_stanza_recursive(stanza, buffer, length);
         if ((size_t)ret > length - 1) {
-            xmpp_free(stanza->ctx, buffer);
+            strophe_free(stanza->ctx, buffer);
             *buf = NULL;
             *buflen = 0;
             return XMPP_EMEM;
@@ -506,10 +507,10 @@ int xmpp_stanza_set_name(xmpp_stanza_t *stanza, const char *name)
         return XMPP_EINVOP;
 
     if (stanza->data)
-        xmpp_free(stanza->ctx, stanza->data);
+        strophe_free(stanza->ctx, stanza->data);
 
     stanza->type = XMPP_STANZA_TAG;
-    stanza->data = xmpp_strdup(stanza->ctx, name);
+    stanza->data = strophe_strdup(stanza->ctx, name);
 
     return stanza->data == NULL ? XMPP_EMEM : XMPP_EOK;
 }
@@ -615,19 +616,19 @@ int xmpp_stanza_set_attribute(xmpp_stanza_t *stanza,
         return XMPP_EINVOP;
 
     if (!stanza->attributes) {
-        stanza->attributes = hash_new(stanza->ctx, 8, xmpp_free);
+        stanza->attributes = hash_new(stanza->ctx, 8, strophe_free);
         if (!stanza->attributes)
             return XMPP_EMEM;
     }
 
-    val = xmpp_strdup(stanza->ctx, value);
+    val = strophe_strdup(stanza->ctx, value);
     if (!val) {
         return XMPP_EMEM;
     }
 
     rc = hash_add(stanza->attributes, key, val);
     if (rc < 0) {
-        xmpp_free(stanza->ctx, val);
+        strophe_free(stanza->ctx, val);
         return XMPP_EMEM;
     }
 
@@ -728,8 +729,8 @@ int xmpp_stanza_set_text(xmpp_stanza_t *stanza, const char *text)
     stanza->type = XMPP_STANZA_TEXT;
 
     if (stanza->data)
-        xmpp_free(stanza->ctx, stanza->data);
-    stanza->data = xmpp_strdup(stanza->ctx, text);
+        strophe_free(stanza->ctx, stanza->data);
+    stanza->data = strophe_strdup(stanza->ctx, text);
 
     return stanza->data == NULL ? XMPP_EMEM : XMPP_EOK;
 }
@@ -758,8 +759,8 @@ int xmpp_stanza_set_text_with_size(xmpp_stanza_t *stanza,
     stanza->type = XMPP_STANZA_TEXT;
 
     if (stanza->data)
-        xmpp_free(stanza->ctx, stanza->data);
-    stanza->data = xmpp_alloc(stanza->ctx, size + 1);
+        strophe_free(stanza->ctx, stanza->data);
+    stanza->data = strophe_alloc(stanza->ctx, size + 1);
     if (!stanza->data)
         return XMPP_EMEM;
 
@@ -844,6 +845,82 @@ const char *xmpp_stanza_get_from(xmpp_stanza_t *stanza)
     return xmpp_stanza_get_attribute(stanza, "from");
 }
 
+/** Get the first child of stanza following a path-like list of names.
+ *  This function searches the children and their children that match
+ *  the given path.
+ *
+ *  * "name" - Search 'name'
+ *
+ *  * "name[@ns='foo']" - Search 'name' which is in the namespace 'foo'
+ *
+ *  The Syntax to pass namespaces is inspired by the XPATH way of passing
+ *  attributes.
+ *
+ *  The namespace syntax only supports single quotes `'`.
+ *
+ *  The \ref XMPP_STANZA_NAME_IN_NS macro is provided as a helper for names
+ *  in namespaces.
+ *
+ *  @param stanza a Strophe stanza object
+ *  @param ... a var-args list that must be terminated by a NULL entry
+ *
+ *  @return the matching child stanza object or NULL if no match was found
+ *
+ *  @ingroup Stanza
+ */
+xmpp_stanza_t *xmpp_stanza_get_child_by_path(xmpp_stanza_t *stanza, ...)
+{
+    xmpp_stanza_t *child = NULL;
+    char *p, *tok, *attr, *saveattr, *ns = NULL;
+    const char *xmlns;
+    va_list ap;
+
+    va_start(ap, stanza);
+
+    while ((p = va_arg(ap, char *)) != NULL) {
+        tok = strophe_strdup(stanza->ctx, p);
+        if (!tok) {
+            child = NULL;
+            break;
+        }
+        saveattr = ns = NULL;
+        attr = strophe_strtok_r(tok, "[", &saveattr);
+        if (attr) {
+            attr = strophe_strtok_r(NULL, "]", &saveattr);
+            if (attr) {
+                if (!strncmp(attr, "@ns='", 5)) {
+                    ns = attr + 5;
+                    strophe_strtok_r(ns, "'", &saveattr);
+                }
+            }
+        }
+        if (!child) {
+            if (strcmp(xmpp_stanza_get_name(stanza), tok))
+                goto error_out;
+
+            if (ns) {
+                xmlns = xmpp_stanza_get_ns(stanza);
+                if (!xmlns || strcmp(xmlns, ns))
+                    goto error_out;
+            }
+            child = stanza;
+        } else {
+            if (!ns)
+                child = xmpp_stanza_get_child_by_name(child, tok);
+            else
+                child = xmpp_stanza_get_child_by_name_and_ns(child, tok, ns);
+        }
+error_out:
+        strophe_free(stanza->ctx, tok);
+        if (!child)
+            break;
+    }
+
+    va_end(ap);
+
+    return p == NULL ? child : NULL;
+}
+
 /** Get the first child of stanza with name.
  *  This function searches all the immediate children of stanza for a child
  *  stanza that matches the name.  The first matching child is returned.
@@ -977,7 +1054,7 @@ char *xmpp_stanza_get_text(xmpp_stanza_t *stanza)
 
     if (stanza->type == XMPP_STANZA_TEXT) {
         if (stanza->data)
-            return xmpp_strdup(stanza->ctx, stanza->data);
+            return strophe_strdup(stanza->ctx, stanza->data);
         else
             return NULL;
     }
@@ -990,7 +1067,7 @@ char *xmpp_stanza_get_text(xmpp_stanza_t *stanza)
     if (len == 0)
         return NULL;
 
-    text = (char *)xmpp_alloc(stanza->ctx, len + 1);
+    text = (char *)strophe_alloc(stanza->ctx, len + 1);
     if (!text)
         return NULL;
 
@@ -1137,8 +1214,8 @@ int xmpp_stanza_del_attribute(xmpp_stanza_t *stanza, const char *name)
 }
 
 /** Create a stanza object in reply to another.
- *  This function makes a copy of a stanza object with the attribute “to” set
- *  its original “from”.
+ *  This function makes a copy of a stanza object with the attribute "to" set
+ *  its original "from".
  *  The stanza will have a reference count of one, so the caller does not
  *  need to clone it.
  *
@@ -1165,7 +1242,7 @@ xmpp_stanza_t *xmpp_stanza_reply(xmpp_stanza_t *stanza)
     copy->type = stanza->type;
 
     if (stanza->data) {
-        copy->data = xmpp_strdup(stanza->ctx, stanza->data);
+        copy->data = strophe_strdup(stanza->ctx, stanza->data);
         if (!copy->data)
             goto copy_error;
     }
@@ -1195,7 +1272,7 @@ copy_error:
  *  Check https://tools.ietf.org/html/rfc6120#section-8.3 for details.
  *
  *  @param stanza a Strophe stanza object
- *  @param error_type type attribute in the <error/> child element
+ *  @param error_type type attribute in the `<error/>` child element
  *  @param condition the defined-condition (e.g. "item-not-found")
  *  @param text optional description, may be NULL
  *
@@ -1310,7 +1387,7 @@ static xmpp_stanza_t *_stanza_new_with_attrs(xmpp_ctx_t *ctx,
     return stanza;
 }
 
-/** Create a <message/> stanza object with given attributes.
+/** Create a `<message/>` stanza object with given attributes.
  *  Attributes are optional and may be NULL.
  *
  *  @param ctx a Strophe context object
@@ -1330,13 +1407,13 @@ xmpp_stanza_t *xmpp_message_new(xmpp_ctx_t *ctx,
     return _stanza_new_with_attrs(ctx, "message", type, id, to);
 }
 
-/** Get text from <body/> child element.
+/** Get text from `<body/>` child element.
  *  This function returns new allocated string. The caller is responsible
  *  for freeing this string with xmpp_free().
  *
- *  @param msg well formed <message/> stanza
+ *  @param msg well formed `<message/>` stanza
  *
- *  @return allocated string or NULL on failure (no <body/> element or
+ *  @return allocated string or NULL on failure (no `<body/>` element or
  *      memory allocation error)
  *
  *  @ingroup Stanza
@@ -1355,9 +1432,10 @@ char *xmpp_message_get_body(xmpp_stanza_t *msg)
     return text;
 }
 
-/** Add <body/> child element to a <message/> stanza with the given text.
+/** Add `<body/>` child element to a `<message/>` stanza with the given text.
  *
- *  @param msg a <message> stanza object without <body/> child element.
+ *  @param msg a `<message>` stanza object without `<body/>` child element.
+ *  @param text The text that shall be placed in the body.
  *
  *  @return 0 on success (XMPP_EOK), and a number less than 0 on failure
  *      (XMPP_EMEM, XMPP_EINVOP)
@@ -1372,7 +1450,7 @@ int xmpp_message_set_body(xmpp_stanza_t *msg, const char *text)
     const char *name;
     int ret;
 
-    /* check that msg is a <message/> stanza and doesn't contain <body/> */
+    /* check that msg is a `<message/>` stanza and doesn't contain `<body/>` */
     name = xmpp_stanza_get_name(msg);
     body = xmpp_stanza_get_child_by_name(msg, "body");
     if (!name || strcmp(name, "message") != 0 || body)
@@ -1399,7 +1477,7 @@ int xmpp_message_set_body(xmpp_stanza_t *msg, const char *text)
     return ret;
 }
 
-/** Create an <iq/> stanza object with given attributes.
+/** Create an `<iq/>` stanza object with given attributes.
  *  Attributes are optional and may be NULL.
  *
  *  @param ctx a Strophe context object
@@ -1415,7 +1493,7 @@ xmpp_stanza_t *xmpp_iq_new(xmpp_ctx_t *ctx, const char *type, const char *id)
     return _stanza_new_with_attrs(ctx, "iq", type, id, NULL);
 }
 
-/** Create a <presence/> stanza object.
+/** Create a `<presence/>` stanza object.
  *
  *  @param ctx a Strophe context object
  *
@@ -1432,7 +1510,7 @@ xmpp_stanza_t *xmpp_presence_new(xmpp_ctx_t *ctx)
  *  The error text is optional and may be NULL.
  *
  *  @param ctx a Strophe context object
- *  @param type enum of xmpp_error_type_t
+ *  @param type enum of strophe_error_type_t
  *  @param text content of a 'text'
  *
  *  @return a new Strophe stanza object
@@ -1527,8 +1605,7 @@ xmpp_error_new(xmpp_ctx_t *ctx, xmpp_error_type_t type, const char *text)
     }
 
     xmpp_stanza_set_ns(error_type, XMPP_NS_STREAMS_IETF);
-    xmpp_stanza_add_child(error, error_type);
-    xmpp_stanza_release(error_type);
+    xmpp_stanza_add_child_ex(error, error_type, 0);
 
     if (text) {
         xmpp_stanza_t *error_text = xmpp_stanza_new(ctx);
@@ -1538,11 +1615,9 @@ xmpp_error_new(xmpp_ctx_t *ctx, xmpp_error_type_t type, const char *text)
         xmpp_stanza_set_ns(error_text, XMPP_NS_STREAMS_IETF);
 
         xmpp_stanza_set_text(content, text);
-        xmpp_stanza_add_child(error_text, content);
-        xmpp_stanza_release(content);
+        xmpp_stanza_add_child_ex(error_text, content, 0);
 
-        xmpp_stanza_add_child(error, error_text);
-        xmpp_stanza_release(error_text);
+        xmpp_stanza_add_child_ex(error, error_text, 0);
     }
 
     return error;
@@ -1563,8 +1638,11 @@ static void _stub_stream_end(char *name, void *userdata)
 
 static void _stream_stanza(xmpp_stanza_t *stanza, void *userdata)
 {
-    stanza = xmpp_stanza_clone(stanza);
-    *(xmpp_stanza_t **)userdata = stanza;
+    xmpp_stanza_t **dest = userdata;
+    if (*dest == NULL) {
+        stanza = xmpp_stanza_clone(stanza);
+        *dest = stanza;
+    }
 }
 
 /** Create a stanza object from the string.
